# -*- coding: utf-8 -*-
# Copyright 2017 - 2022 Avram Lubkin, All Rights Reserved

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
Test module for enlighten._util
"""

from textwrap import dedent

import blessed

from enlighten._util import format_time, Lookahead, HTMLConverter

from tests import TestCase, MockTTY


class TestFormatTime(TestCase):
    """
    Test cases for :py:func:`_format_time`
    """

    def test_seconds(self):
        """Verify seconds formatting"""

        self.assertEqual(format_time(0), '00:00')
        self.assertEqual(format_time(6), '00:06')
        self.assertEqual(format_time(42), '00:42')

    def test_minutes(self):
        """Verify minutes formatting"""

        self.assertEqual(format_time(60), '01:00')
        self.assertEqual(format_time(128), '02:08')
        self.assertEqual(format_time(1684), '28:04')

    def test_hours(self):
        """Verify hours formatting"""

        self.assertEqual(format_time(3600), '1h 00:00')
        self.assertEqual(format_time(43980), '12h 13:00')
        self.assertEqual(format_time(43998), '12h 13:18')

    def test_days(self):
        """Verify days formatting"""

        self.assertEqual(format_time(86400), '1d 0h 00:00')
        self.assertEqual(format_time(1447597), '16d 18h 06:37')


class TestLookahead(TestCase):
    """
    Test cases for Lookahead
    """

    def test_iteration(self):
        """Verify normal iteration"""

        wrapped = Lookahead(iter(range(10)))

        self.assertEqual([next(wrapped) for _ in range(10)], list(range(10)))

    def test_getitem(self):
        """Verify __getitem__ behavior"""

        wrapped = Lookahead(iter(range(10)))

        self.assertEqual(wrapped[0], 0)
        self.assertEqual(wrapped[4], 4)
        self.assertEqual(wrapped[2: 4], [2, 3])
        self.assertEqual(wrapped[8: 12], [8, 9])

        with self.assertRaisesRegex(TypeError, 'Index or slice notation is required'):
            wrapped['named_key']  # pylint: disable=pointless-statement

        with self.assertRaisesRegex(ValueError, 'Negative indexes are not supported'):
            wrapped[-1]  # pylint: disable=pointless-statement

    def test_buffer(self):
        """Output changes as iteration proceeds"""

        wrapped = Lookahead(iter(range(10)))

        self.assertEqual(next(wrapped), 0)
        self.assertEqual(wrapped[0], 1)
        self.assertEqual(next(wrapped), 1)
        self.assertEqual(wrapped[4], 6)
        self.assertEqual(next(wrapped), 2)
        self.assertEqual(wrapped[2: 4], [5, 6])
        self.assertEqual(next(wrapped), 3)
        self.assertEqual(wrapped[8: 12], [])

    def test_step_notation(self):
        """Slice notation is supported"""

        wrapped = Lookahead(iter(range(10)))

        self.assertEqual(wrapped[: 6: 2], [0, 2, 4])


class TestHTMLConverter(TestCase):
    """
    Test cases for HTMLConverter
    """

    # pylint: disable=protected-access

    @classmethod
    def setUpClass(cls):
        cls.tty = MockTTY()
        cls.term = blessed.Terminal(
            stream=cls.tty.stdout, force_styling=True
        )
        cls.term.number_of_colors = 1 << 24

    @classmethod
    def tearDownClass(cls):
        cls.tty.close()

    def setUp(self):
        self.converter = HTMLConverter(term=self.term)

    def test_color(self):
        """Verify color conversion"""

        # CGA color on RGB color
        out = self.converter.to_html(self.term.blue_on_aquamarine('blue_on_aquam'))
        self.assertEqual(
            out,
            u'<pre><span class="enlighten-fg-blue enlighten-bg-aquamarine">blue_on_aquam</span></pre>'
        )

        self.assertEqual(self.converter._styles['enlighten-fg-blue'], {'color': '#0000ee'})
        self.assertEqual(
            self.converter._styles['enlighten-bg-aquamarine'], {'background-color': '#7fffd4'}
        )

        # RGB color on CGA color
        out = self.converter.to_html(self.term.aquamarine_on_blue('aquam_on_blue'))
        self.assertEqual(
            out,
            u'<pre><span class="enlighten-fg-aquamarine enlighten-bg-blue">aquam_on_blue</span></pre>'
        )

        self.assertEqual(self.converter._styles['enlighten-fg-aquamarine'], {'color': '#7fffd4'})
        self.assertEqual(
            self.converter._styles['enlighten-bg-blue'], {'background-color': '#0000ee'}
        )

        # On RGB color
        out = self.converter.to_html(self.term.on_color_rgb(80, 4, 13)('on_color_rgb'))
        self.assertEqual(out, '<pre><span class="enlighten-bg-50040d">on_color_rgb</span></pre>')

        self.assertEqual(
            self.converter._styles['enlighten-bg-50040d'], {'background-color': '#50040d'}
        )

        # 256 Color
        out = self.converter.to_html(self.term.color(90)('color_90'))
        self.assertEqual(out, '<pre><span class="enlighten-fg-870087">color_90</span></pre>')

        self.assertEqual(self.converter._styles['enlighten-fg-870087'], {'color': '#870087'})

        # On 256 Color
        out = self.converter.to_html(self.term.on_color(90)('on_color_90'))
        self.assertEqual(out, '<pre><span class="enlighten-bg-870087">on_color_90</span></pre>')

        self.assertEqual(
            self.converter._styles['enlighten-bg-870087'], {'background-color': '#870087'}
        )

        # CGA Bright Color
        out = self.converter.to_html(self.term.bright_red('bright_red'))
        self.assertEqual(out, '<pre><span class="enlighten-fg-bright-red">bright_red</span></pre>')

        self.assertEqual(self.converter._styles['enlighten-fg-bright-red'], {'color': '#ff0000'})

        # On CGA Bright Color
        out = self.converter.to_html(self.term.on_bright_red('on_bright_red'))
        self.assertEqual(
            out,
            '<pre><span class="enlighten-bg-bright-red">on_bright_red</span></pre>'
        )

        self.assertEqual(
            self.converter._styles['enlighten-bg-bright-red'], {'background-color': '#ff0000'}
        )

    def test_style(self):
        """Verify style conversion"""

        # Italics
        out = self.converter.to_html(self.term.italic('italic'))
        self.assertEqual(out, '<pre><span class="enlighten-italic">italic</span></pre>')

        self.assertEqual(self.converter._styles['enlighten-italic'], {'font-style': 'italic'})

        # Bold
        out = self.converter.to_html(self.term.bold('bold'))
        self.assertEqual(out, '<pre><span class="enlighten-bold">bold</span></pre>')

        self.assertEqual(self.converter._styles['enlighten-bold'], {'font-weight': 'bold'})

        # Underline
        out = self.converter.to_html(self.term.underline('underline'))
        self.assertEqual(out, '<pre><span class="enlighten-underline">underline</span></pre>')

        self.assertEqual(
            self.converter._styles['enlighten-underline'], {'text-decoration': 'underline'}
        )

    def test_unsupported(self):
        """Verify unsupported does not produce classes"""

        # Unsupported capability
        out = self.converter.to_html(self.term.move(5, 6) + 'unsupported_move')
        self.assertEqual(out, '<pre>unsupported_move</pre>')

        # Unsupported text attribute
        out = self.converter.to_html(self.term.reverse('unsupported_reverse'))
        self.assertEqual(out, '<pre>unsupported_reverse</pre>')

    def test_link(self):
        """Verify link creates hyperlink"""

        out = self.converter.to_html(
            self.term.link('https://pypi.org/project/enlighten/', 'enlighten')
        )
        self.assertEqual(
            out,
            '<pre><a href="https://pypi.org/project/enlighten/">enlighten<a></pre>'
        )

    def test_empty_span(self):
        """Empty Spans are ignored"""

        out = self.converter.to_html(self.term.underline('') + 'empty')
        self.assertEqual(out, '<pre>empty</pre>')

    def test_class_not_unique(self):
        """Repeated classes are dropped within the same span"""

        out = self.converter.to_html(self.term.blue_on_aquamarine(self.term.blue('blue_on_aquam')))
        self.assertEqual(
            out,
            u'<pre><span class="enlighten-fg-blue enlighten-bg-aquamarine">blue_on_aquam</span></pre>'
        )

    def test_style_output(self):
        """Verify style section output"""

        out = self.converter.to_html(self.term.red_on_slategrey('red_on_slategrey'))

        self.assertEqual(
            out,
            u'<pre><span class="enlighten-fg-red enlighten-bg-slategray">red_on_slategrey</span></pre>'
        )

        style = '''\
        <style>
        .enlighten-fg-red {
          color: #cd0000;
        }
        .enlighten-bg-slategray {
          background-color: #708090;
        }
        </style>
        '''

        self.assertEqual(self.converter.style, dedent(style))

    def test_blink(self):
        """Blink requires an additional style section"""

        if not self.term.blink:
            self.skipTest('blink is not supported by this terminal')

        out = self.converter.to_html(self.term.blink('blink'))
        self.assertEqual(out, '<pre><span class="enlighten-blink">blink</span></pre>')

        self.assertEqual(
            self.converter._additional_styles,
            {'@keyframes enlighten-blink-animation {\n  to {\n    visibility: hidden;\n  }\n}'}
        )

        self.assertEqual(
            self.converter._styles['enlighten-blink'],
            {'animation': 'enlighten-blink-animation 1s steps(5, start) infinite'}
        )

        style = '''\
        <style>
        .enlighten-blink {
          animation: enlighten-blink-animation 1s steps(5, start) infinite;
        }
        @keyframes enlighten-blink-animation {
          to {
            visibility: hidden;
          }
        }
        </style>
        '''

        self.assertEqual(self.converter.style, dedent(style))



# -*- coding: utf-8 -*-
# Copyright 2017 - 2021 Avram Lubkin, All Rights Reserved

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
**Enlighten status bar submodule**

Provides StatusBar class
"""

import time

from enlighten._basecounter import PrintableCounter
from enlighten._util import (EnlightenWarning, FORMAT_MAP_SUPPORT, format_time,
                             Justify, raise_from_none, warn_best_level)


STATUS_FIELDS = {'elapsed', 'fill'}


class StatusBar(PrintableCounter):
    """
    Args:
        enabled(bool): Status (Default: :py:data:`True`)
        color(str): Color as a string or RGB tuple see :ref:`Status Color <status_color>`
        fields(dict): Additional fields used for :ref:`formating <status_format>`
        fill(str): Fill character used in formatting and justifying text (Default: ' ')
        justify(str):
            One of :py:attr:`Justify.CENTER`, :py:attr:`Justify.LEFT`, :py:attr:`Justify.RIGHT`
        leave(True): Leave status bar after closing (Default: :py:data:`True`)
        min_delta(float): Minimum time, in seconds, between refreshes (Default: 0.1)
        status_format(str): Status bar format, see :ref:`Format <status_format>`

    Status bar class

    A :py:class:`StatusBar` instance should be created with the :py:meth:`Manager.status_bar`
    method.

    .. _status_color:

    **Status Color**

    Color works similarly to color on :py:class:`Counter`, except it affects the entire status bar.
    See :ref:`Series Color <series_color>` for more information.

    .. _status_format:

    **Format**

    There are two ways to populate the status bar, direct and formatted. Direct takes
    precedence over formatted.

    .. _status_format_direct:

    **Direct Status**

    Direct status is used when arguments are passed to :py:meth:`Manager.status_bar` or
    :py:meth:`StatusBar.update`. Any arguments are coerced to strings and joined with a space.
    For example:

    .. code-block:: python


        status_bar.update('Hello', 'World!')
        # Example output: Hello World!

        status_bar.update('Hello World!')
        # Example output: Hello World!

        count = [1, 2, 3, 4]
        status_bar.update(*count)
         # Example output: 1 2 3 4

    .. _status_format_formatted:

    **Formatted Status**

        Formatted status uses the format specified in the ``status_format`` parameter to populate
        the status bar.

        .. code-block:: python

            'Current Stage: {stage}'

            # Example output
            'Current Stage: Testing'

        Available fields:

            - elapsed(:py:class:`str`) - Time elapsed since instance was created
            - fill(:py:class:`str`) - Filled with :py:attr:`fill` until line is width of terminal.
              May be used multiple times. Minimum width is 3.

        .. note::

            The status bar is only updated when :py:meth:`StatusBar.update` or
            :py:meth:`StatusBar.refresh` is called, so fields like ``elapsed``
            will need additional calls to appear dynamic.

        User-defined fields:

            Users can define fields in two ways, the ``fields`` parameter and by passing keyword
            arguments to :py:meth:`Manager.status_bar` or :py:meth:`StatusBar.update`

            The ``fields`` parameter can be used to pass a dictionary of additional
            user-defined fields. The dictionary values can be updated after initialization to allow
            for dynamic fields. Any fields that share names with available fields are ignored.

            If fields are passed as keyword arguments to :py:meth:`Manager.status_bar` or
            :py:meth:`StatusBar.update`, they take precedent over the ``fields`` parameter.


    **Instance Attributes**

        .. py:attribute:: elapsed

            :py:class:`float` - Time since start

        .. py:attribute:: enabled

            :py:class:`bool` - Current status

        .. py:attribute:: manager

            :py:class:`Manager` - Manager Instance

        .. py:attribute:: position

            :py:class:`int` - Current position

    """

    __slots__ = ('fields', '_justify', 'status_format', '_static', '_fields')

    def __init__(self, *args, **kwargs):

        super(StatusBar, self).__init__(keywords=kwargs)

        self.fields = kwargs.pop('fields', {})
        self._justify = None
        self.justify = kwargs.pop('justify', Justify.LEFT)
        self.status_format = kwargs.pop('status_format', None)
        self._fields = kwargs
        self._static = ' '.join(str(arg) for arg in args) if args else None

    @property
    def justify(self):
        """
        Maps to justify method determined by ``justify`` parameter
        """
        return self._justify

    @justify.setter
    def justify(self, value):

        if value in (Justify.LEFT, Justify.CENTER, Justify.RIGHT):
            self._justify = getattr(self.manager.term, value)

        else:
            raise ValueError("justify must be one of Justify.LEFT, Justify.CENTER, ",
                             "Justify.RIGHT, not: '%r'" % value)

    def format(self, width=None, elapsed=None):
        """
        Args:
            width (int): Width in columns to make progress bar
            elapsed(float): Time since started. Automatically determined if :py:data:`None`

        Returns:
            :py:class:`str`: Formatted status bar

        Format status bar
        """

        width = width or self.manager.width
        justify = self.justify

        # If static message was given, just return it
        if self._static is not None:
            rtn = self._static

        # If there is no format, return empty
        elif self.status_format is None:
            rtn = ''

        # Generate from format
        else:
            fields = self.fields.copy()
            fields.update(self._fields)

            # Warn on reserved fields
            reserved_fields = (set(fields) & STATUS_FIELDS)
            if reserved_fields:
                warn_best_level('Ignoring reserved fields specified as user-defined fields: %s' %
                                ', '.join(reserved_fields),
                                EnlightenWarning)

            elapsed = elapsed if elapsed is not None else self.elapsed
            fields['elapsed'] = format_time(elapsed)
            fields['fill'] = self._placeholder_

            # Format
            try:
                if FORMAT_MAP_SUPPORT:
                    rtn = self.status_format.format_map(fields)
                else:  # pragma: no cover
                    rtn = self.status_format.format(**fields)
            except KeyError as e:
                raise_from_none(ValueError('%r specified in format, but not provided' % e.args[0]))

        rtn = self._fill_text(rtn, width)

        return self._colorize(justify(rtn, width=width, fillchar=self.fill))

    def update(self, *objects, **fields):  # pylint: disable=arguments-differ
        """
        Args:
            objects(list): Values for :ref:`Direct Status <status_format_direct>`
            force(bool): Force refresh even if ``min_delta`` has not been reached
            fields(dict): Fields for for :ref:`Formatted Status <status_format_formatted>`

        Update status and redraw

        Status bar is only redrawn if ``min_delta`` seconds past since the last update
        """

        force = fields.pop('force', False)

        self._static = ' '.join(str(obj) for obj in objects) if objects else None
        self._fields.update(fields)

        if self.enabled:
            currentTime = time.time()
            if force or currentTime - self.last_update >= self.min_delta:
                self.refresh(elapsed=currentTime - self.start)



# -*- coding: utf-8 -*-
# Copyright 2017 - 2023 Avram Lubkin, All Rights Reserved

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
**Enlighten manager submodule**

Provides Manager class
"""

from IPython.display import DisplayHandle, HTML

from enlighten._basemanager import BaseManager
from enlighten._util import HTMLConverter


class NotebookManager(BaseManager):
    """
    Args:
        counter_class(:py:term:`class`): Progress bar class (Default: :py:class:`Counter`)
        status_bar_class(:py:term:`class`): Status bar class (Default: :py:class:`StatusBar`)
        enabled(bool): Status (Default: True)
        width(int): Static output width (Default: 100)
        kwargs(Dict[str, Any]): Any additional :py:term:`keyword arguments<keyword argument>`
            will be used as default values when :py:meth:`counter` is called.

    Manager class for outputting progress bars to Jupyter notebooks

    The following keyword arguments are set if provided, but ignored:

      * *stream*
      * *set_scroll*
      * *companion_stream*
      * *no_resize*
      * *threaded*

    """

    def __init__(self, **kwargs):

        # Force terminal to xterm-256color because it should have broad support
        kwargs['term'] = 'xterm-256color'

        super(NotebookManager, self).__init__(**kwargs)

        # Force 24-bit color
        self.term.number_of_colors = 1 << 24

        self._converter = HTMLConverter(self.term)
        self._output = []
        self._display = DisplayHandle()
        self._html = HTML('')
        self._primed = False

        # Default width to 100 unless specified
        self.width = self._width or 100

    def __repr__(self):
        return '%s()' % self.__class__.__name__

    def _flush_streams(self):
        """
        Display buffered output
        """

        if not self.enabled:
            return

        self._html.data = '%s<div class="enlighten">\n%s\n</div>\n' % (
                          self._converter.style, '\n'.join(reversed(self._output)))

        if self._primed:
            self._display.update(self._html)
        else:
            self._primed = True
            self._display.display(self._html)

    def stop(self):
        # See parent class for docstring

        if not self.enabled:
            return

        positions = self.counters.values()

        if positions:
            for num in range(max(positions), 0, -1):
                if num not in positions:
                    self._output[num - 1] = '  <br>'

        for counter in self.counters:
            counter.enabled = False

        self._flush_streams()

    def write(self, output='', flush=True, counter=None, **kwargs):
        # See parent class for docstring

        if not self.enabled:
            return

        position = self.counters[counter] if counter else 1

        # If output is callable, call it with supplied arguments
        if callable(output):
            output = output(**kwargs)

        # If there is space between this bar and the last, fill with blank lines
        for _ in range(position - len(self._output)):
            self._output.append('  <br>')

        # Set output
        self._output[position - 1] = (
            '  <div class="enlighten-bar">\n    %s\n  </div>' % self._converter.to_html(output)
        )

        if flush:
            self._flush_streams()



# -*- coding: utf-8 -*-
# Copyright 2017 - 2022 Avram Lubkin, All Rights Reserved

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
**Enlighten manager submodule**

Provides Manager classes and utilities
"""

import sys

from enlighten._counter import Counter
from enlighten._manager import Manager


try:
    from IPython import get_ipython
    IN_NOTEBOOK = 'IPKernelApp' in get_ipython().config
    from enlighten._notebook_manager import NotebookManager  # pylint: disable=ungrouped-imports
except (ImportError, AttributeError):
    IN_NOTEBOOK = False


def get_manager(stream=None, counter_class=Counter, **kwargs):
    """
    Args:
        stream(:py:term:`file object`): Output stream. If :py:data:`None`,
            defaults to :py:data:`sys.__stdout__`
        counter_class(:py:term:`class`): Progress bar class (Default: :py:class:`Counter`)
        kwargs(Dict[str, Any]): Any additional :py:term:`keyword arguments<keyword argument>`
            will passed to the manager class.

    Returns:
        :py:class:`Manager`: Manager instance

    Convenience function to get a manager instance

    If running inside a notebook, a :py:class:`NotebookManager`
    instance is returned. otherwise a standard :py:class:`Manager` instance is returned.

    If a a standard :py:class:`Manager` instance is used and ``stream`` is not attached
    to a TTY, the :py:class:`Manager` instance is disabled.
    """

    if IN_NOTEBOOK:
        return NotebookManager(stream=stream, counter_class=counter_class, **kwargs)

    stream = sys.__stdout__ if stream is None else stream
    isatty = hasattr(stream, 'isatty') and stream.isatty()
    kwargs['enabled'] = isatty and kwargs.get('enabled', True)
    return Manager(stream=stream, counter_class=counter_class, **kwargs)
